/****************************************************************************
 *                                                                          *
 * File    : main.c                                                         *
 *                                                                          *
 * Purpose : Console mode (command line) program.                           *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

#include <stdio.h>
//#include <mmsystem.h>
#include <stdlib.h>

// Global declarations
#define TEMPLATE_MELODY_LEN	70
#define TIME_OFFSET	0.f
#define SAMPLE_FREQUENCY		48000
#define SAMPLE_FREQUENCY_01	 (SAMPLE_FREQUENCY/100)
#define BPM					 70
#define ROW_COUNT			16 * TEMPLATE_MELODY_LEN
#define PLAY_TIME				ROW_COUNT * 60 / ( BPM * 4 )
#define ROW_SIZE_SAMPLES	SAMPLE_FREQUENCY * 60 / ( BPM * 4 )
#define MASTER_VOLUME		0.51f
#define ROW_LEN_SEC		60.0f / ( BPM * 4 )
#define PI 3.1415926535897932384626433832795f
#define PI_2  6.283184f


struct wavehdr44
{
char wavefile[4];			//	db	'RIFF'
long	size36;				//	dd 	0;wavefilesize-8;datasize+36;
char	wavef[8];				//	db	'WAVEfmt '
long	bits16;				//	dd	16
short	word1;				//		dw	1
short	channels;			//dw	chnls		
long	freq;			//dd	freq
long	Bytespersecond;		//dd	freq*chnls*bytesperchnl
short	Block_alignment;		//dw	chnls*bytesperchnl
short	Bits_per_sample;		//dw	bits
char dta[4];					//db	'data'
long	Sizeofdata;			//dd	0;datasize
}ldwhdr,wavhdr={{'R','I','F','F'},
			0,
		{'W','A','V','E','f','m','t',' '},
			16,1,
			2,
			SAMPLE_FREQUENCY,
		(SAMPLE_FREQUENCY*4),
		4,
			16,
		{'d','a','t','a'},
			0			};



//HWAVEOUT	g_wo;
int			g_patternNo;
int			g_patternRowNo;

struct Instrument { 
	char note;
	char noteEnd;
	char noteLen;
	char volume;
	char oscType;
	char noteOffset;
	char attackTime;
	char decayTime;
	char sustainTime;
	char filterCutoff;
	char filterCutoffModulation;
	char filterRes;
	char echoDelay;
	char echoRepeats;
	char echoDamping;
	char wrapLevel;
	char hardLevel;
	char compressorLow;
	char compressorHigh;
	char pan;
	char resFilterFreq;
	char resFilterFreqModulation;
	char resFilterRes;
};

struct Instrument g_instruments[] = { { -1,0,61,25,1,-5,0,37,47,18,-2,58,10,3,50,0,0,7,25,0,23,2,68 }
, { -2,0,26,50,2,-20,0,45,80,0,0,0,5,3,61,13,0,0,0,16,40,7,12 }
, { -4,0,20,38,0,-30,0,11,28,0,0,0,0,0,0,0,0,0,0,0,0,0,0 }
, { -3,0,9,27,3,0,0,36,36,47,0,51,10,10,33,0,0,0,0,0,96,0,69 }
, { -3,10,9,36,0,-20,0,15,29,108,0,24,0,0,0,0,0,22,99,0,42,0,79 }
, { -4,24,12,79,0,-6,0,28,127,0,0,0,0,0,0,0,0,0,0,0,0,0,0 }
, { -5,0,21,50,1,-20,10,20,80,0,0,0,10,1,25,48,0,0,0,0,-6,5,68 }
, { -7,0,3,19,2,-5,0,60,60,-6,0,61,5,100,90,0,0,0,0,0,0,0,0 }
, { -5,0,4,-3,0,-33,0,127,127,4,0,56,9,11,52,0,0,35,91,16,8,0,72 }
, { -8,0,5,50,2,-11,14,127,127,0,0,0,8,1,67,40,0,0,0,0,0,0,0 }
, { -8,0,16,50,1,-18,0,0,0,0,0,0,4,2,56,0,0,0,0,0,0,0,0 }
, { -8,0,11,50,0,-40,0,0,0,0,0,0,0,0,0,0,0,0,84,0,0,0,0 }
 };


const char g_bytePatterns[][16] = { {45,0,52,0,48,0,45,0,52,0,48,0,45,57,52,64},
{47,0,54,0,51,0,47,0,54,0,51,0,47,0,54,0},
{45,0,67,0,48,0,45,0,52,0,63,0,45,57,67,64},
{47,0,69,0,51,0,47,0,54,0,66,62,47,59,69,0},
{33,0,0,33,0,0,33,0,34,0,34,0,34,34,0,0},
{38,38,0,0,38,0,38,0,39,0,39,0,39,0,0,0},
{73,0,0,0,0,0,0,0,76,0,0,0,0,0,0,0},
{78,0,0,76,74,0,76,0,73,0,74,0,71,0,73,0},
{69,0,0,0,0,0,0,68,69,0,0,0,0,0,0,66},
{68,0,0,0,0,0,0,65,66,0,68,0,69,0,71,0},
{73,0,0,0,73,0,0,0,74,73,0,71,0,73,0,73},
{76,0,74,0,73,0,71,0,73,71,69,71,69,68,69,68},
{66,0,0,0,0,65,0,66,68,0,0,0,0,68,0,69},
{71,69,0,71,73,71,0,74,73,76,0,76,74,73,0,71},
{0,0,45,0,0,0,45,0,0,0,45,0,0,0,45,45},
{45,0,0,0,45,0,0,0,45,0,0,0,45,0,0,0},
{0,33,0,33,0,33,0,33,0,33,0,33,0,33,0,33},
{45,57,45,57,45,57,45,57,48,60,48,60,47,59,47,59},
{45,57,45,57,45,57,45,57,51,63,51,63,48,60,48,60},
{45,57,45,57,45,57,45,57,49,61,49,61,52,64,52,64},
{109,109,109,109,109,109,109,109,109,109,109,109,109,109,109,109},
{109,103,96,90,83,76,70,63,57,50,43,37,30,24,17,10},
{10,16,23,29,36,43,49,56,62,69,76,82,89,95,102,109},
{45,0,45,45,0,0,0,45,0,45,45,0,0,0,0,0},
{0,0,0,0,33,45,33,45,0,0,0,0,0,0,0,0},
{0,0,0,0,33,0,33,0,0,0,0,0,33,0,0,0},
{0,36,0,36,0,36,0,36,0,36,0,36,0,36,0,36},
{0,60,0,60,0,60,0,60,0,60,0,60,0,60,0,60},
{45,0,45,0,0,45,0,0,45,0,45,0,0,0,0,0},
{45,61,48,45,61,48,45,61,48,45,61,48,45,61,48,60},
{46,60,49,46,60,49,46,60,49,46,60,49,46,60,49,63},
{45,49,61,45,49,61,45,49,64,45,49,64,45,49,64,60},
{49,61,63,49,61,63,49,61,63,49,61,60,49,61,63,60}};

const char g_byteSequences[][TEMPLATE_MELODY_LEN] = { {0,0,0,0,1,2,3,4,1,2,3,4,1,2,3,4,3,4,3,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,7,8,9,10,11,12,13,14,7,8,9,10,11,12,13,14,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
{17,17,17,17,17,17,17,17,5,6,5,6,5,6,5,6,0,0,0,0,17,17,17,17,28,28,27,27,17,17,17,17,0,0,0,0,0,0,17,17,17,17,27,17,17,17,17,17,17,17,17,17,17,17,0,0,0,0,25,25,25,25,0,0,0,0,0,0,0,0},
{15,15,15,26,15,15,15,15,15,15,15,15,15,15,15,15,0,0,0,0,15,15,15,15,15,15,15,15,0,0,0,0,0,0,0,0,0,0,0,0,0,0,15,15,15,15,15,15,0,0,15,15,15,15,0,0,0,0,26,26,26,26,25,15,25,25,25,25,0,0},
{16,16,16,24,16,16,16,16,16,16,16,16,16,16,16,16,0,0,0,0,16,16,16,16,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16,16,16,16,16,16,0,0,16,16,16,16,0,0,0,0,24,24,24,24,24,29,24,24,29,24,0,0},
{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,18,20,18,20,18,20,18,20,18,20,18,20,18,19,18,20,0,0,0,0,0,0,0,0,0,0,0,0,0,0,18,20,18,20,5,6,5,6,17,17,17,17,0,0,0,0,0,0,0,0},
{0,0,0,0,0,0,0,0,23,21,21,22,0,0,0,0,0,0,0,0,21,21,21,21,21,22,0,0,0,0,23,21,21,0,21,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,21,21,21,23,0,21,23,22,0,0,23,22,0,0,0,0,0,0,0,0},
{0,0,0,0,0,0,0,0,1,2,3,4,1,2,3,4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,30,31,32,33,30,31,32,33}};

int INSTRUMENT_COUNT = sizeof(g_instruments) / sizeof(g_instruments[0]);

// Some auxiliary functions
float Sign ( float f ) {
	if ( f > 0.0f )
		return 1.0f;
	return -1.0f;
}
float CutLevel ( float x, float lvl ) {
	if ( x > lvl )
		return lvl;
	if ( x < -lvl )
		return -lvl;
	return x;
}
float buf0LPRes, buf1LPRes;
float LowPassResonantFilter ( float x, float f, float q ) {
	float fb = q + q / ( 1.0f - f );
	buf0LPRes = buf0LPRes + f * ( x - buf0LPRes + fb * ( buf0LPRes - buf1LPRes ) );
	buf1LPRes = buf1LPRes + f * ( buf0LPRes - buf1LPRes );
	return buf1LPRes;
}
float buf1Res, buf2Res;
float my_sin ( float v ) {
	__asm { 
		fld v
		fsin
		fstp v
	}
	return v;
}
float my_cos ( float v ) {
	__asm {
		fld v
		fcos
		fstp v
	}
	return v;
}
float my_sqrt ( float v ) {
	__asm {
		fld v
		fsqrt
		fstp v
	}
	return v;
}
float my_fabs ( float v ) {
	__asm {
		fld v
		fabs
		fstp v
	}
	return v;
}
int _ftol ( float fvar ) {
	int lvar;
	__asm {
		fld fvar
		fistp lvar
	}
	return lvar;
}
float pow2 ( float x ) {
	float y;
	__asm {
		fld dword ptr x
		fld st(0)
		frndint
		fxch st(1)
		fsub st(0),st(1)
		f2xm1
		fld1
		faddp st(1),st(0)
		fscale
		fstp st(1)
		fstp dword ptr y
	}
	return y;
}
float Noise1 ( int x ) {
    x = (x<<13) ^ x;
    return (float)( 1.0f - ( (x * (x * x * 15731 + 789221) + 1376312589) & 0x7fffffff) / 1073741824.0f);    
}
float ResonantFilter ( float x, float f, float q ) {
	float y = x + 2.0f * my_cos ( PI * f * 0.25f ) * buf1Res * q - buf2Res * q * q;
	buf2Res = buf1Res;
	buf1Res = y;
	return buf1Res;
}
float Sawtooth ( float x ) {
	return ( ( x - _ftol ( x / PI_2 ) * PI_2 ) / PI_2 ) * 2.0f;
}
float NoteToFrequency ( char note  ) {
	return PI_2 * 440.0f * pow2 ( (note-45)/12.0f );
}
float Square ( float x ) {
	return Sign ( Sawtooth ( x ) );
}
char GetFixedOrStreamValue ( char val ) {
	if ( val >= 0 )
		return val;
	val = -val;
	int k = g_byteSequences[val-1][g_patternNo];
	if ( k )
		return g_bytePatterns[k-1][g_patternRowNo];
	return 0;
}

//WAVEFORMATEX wfx = { WAVE_FORMAT_PCM, 2, SAMPLE_FREQUENCY, SAMPLE_FREQUENCY*4, 4, 16, 0 };
short finalBuf[(PLAY_TIME+20)*SAMPLE_FREQUENCY*2]  ; // We need +20 since some sounds (e.g. echoed) can last for a long time. Hope it's large enough :)
float data[(PLAY_TIME+20)*SAMPLE_FREQUENCY*2] ;
//WAVEHDR hdr = {0};

// Creates large buffer and fills it with samples. Starts playing generated sound.
// Pefroms no error checking and does not dispose resources.
void makefile ( ) {
//	waveOutOpen ( &g_wo, WAVE_MAPPER, &wfx, NULL, 0, CALLBACK_NULL );
	// Loop through all instruments and create sound for each one of them independently
	for ( int i = 0; i < INSTRUMENT_COUNT; i++ ) {
		printf(".");
		float *b = data;
		struct Instrument* ins = g_instruments + i; 
		int oscType = ins->oscType;
		float attackTime = ins->attackTime * 0.01f;
		float decayTime = ins->decayTime * 0.01f;
		float sustainTime = ins->sustainTime * 0.01f;
		float tt = ins->pan*0.01f; float pl = 1.0f - tt; float pr = 1.0f + tt;

		// Loop through all rows for given instrument
		for ( int r = 0; r < ROW_COUNT; r++, b += ROW_SIZE_SAMPLES*2 ) {
			g_patternNo = ( r >> 4 );
			g_patternRowNo = ( r & 0xF );
			char note = GetFixedOrStreamValue ( ins->note );
			if ( !note )
				continue;
			note += ins->noteOffset;
			int noteLenSamples = ins->noteLen * SAMPLE_FREQUENCY_01;
			float f = NoteToFrequency ( note );
			float f1 = f;
			if ( ins->noteEnd )
				f1 = NoteToFrequency ( ins->noteEnd + ins->noteOffset );
			float filtF = GetFixedOrStreamValue ( ins->filterCutoff ) * 0.01f;
			float filtQ = ins->filterRes * 0.01f;
			float filtMod = ins->filterCutoffModulation * 0.01f;
			float resF = GetFixedOrStreamValue ( ins->resFilterFreq ) * 0.01f;
			float resQ = ins->resFilterRes * 0.01f;
			float resMod = ins->resFilterFreqModulation * 0.01f;
			buf0LPRes = buf1LPRes = 0; buf1Res = buf2Res = 0;
			float vol = GetFixedOrStreamValue ( ins->volume ) * 0.01f;
			float wrapLevel = ins->wrapLevel * 0.01f;
			float hardLevel = ins->hardLevel * 0.01f;
			float compLow = ins->compressorLow * 0.01f;
			float compHigh = ins->compressorHigh * 0.01f;
			float compDL = compHigh - compLow;
			float *b1 = b;

			// Create one sound sample based on instrument and sequence parameters
			for ( int k = 0; k < noteLenSamples; k++, b1 += 2 ) {
				float t1 = (float) k / noteLenSamples;
				float t = (float) k / SAMPLE_FREQUENCY * ( f + (f1-f)*t1 );
				float y;
				if ( oscType == 0 )
					y = my_sin ( t );
				if ( oscType == 1 )
					y = Sawtooth ( t );
				if ( oscType == 2 )
					y = Square ( t );
				if ( oscType == 3 )
					y = y = Noise1 ( k ) - 0.5f;
				if ( t1 < attackTime )
					y *= t1 / attackTime;
				else if ( t1 < decayTime )
					y *= 1.0f - 0.3f * ( t1 - attackTime ) / ( decayTime - attackTime );
				else if ( t1 < sustainTime )
					y *= 0.7f;
				else
					y *= 0.7f * ( 1.0f - t1  ) / ( 1.0f - sustainTime );
				if ( filtF )
					y = LowPassResonantFilter ( y, filtF + filtMod * t1, filtQ );
				if ( resF )
					y = ResonantFilter ( y, resF + resMod * t1, resQ );
				if ( wrapLevel )
					while ( my_fabs ( y ) > wrapLevel )
						y = 2.0f * wrapLevel * Sign ( y ) - y;
				if ( compHigh ) {
					float v = my_fabs ( y ) - compLow;
					if ( v > 0.0f ) {
						v = compHigh - compDL / ( 1.0f + 10.0f * v / compDL );
						y = v * Sign ( y );
					}
				}
				float ss = y * vol;
				*b1 += ss*pl;
				*(b1+1) += ss*pr;
				if ( ins->echoRepeats ) {
					float c = ins->echoDamping*0.01f;
					for ( int kk = 0; kk < ins->echoRepeats; kk++ ) {
						float *ad = b1+_ftol((kk+1)*ins->echoDelay*(0.01f*SAMPLE_FREQUENCY*2));
						*(ad++) += ss*pl*c;
						*ad += ss*pr*c;
						c *= c;
					}
				}
			}
		}
	}

	// Copy just generated sound to audio buffer and play it
	short *s = finalBuf;
	float *d = data;
	for ( int i = 0; i < 2*PLAY_TIME*SAMPLE_FREQUENCY; i++ )
		*(s++) = _ftol ( 32000.0f * CutLevel ( (*(d++)) * MASTER_VOLUME, 1 ) );
//	hdr.lpData = (LPSTR) ( finalBuf + 2*(int)(TIME_OFFSET*SAMPLE_FREQUENCY));
//	hdr.dwBufferLength  = (PLAY_TIME-TIME_OFFSET)*SAMPLE_FREQUENCY*sizeof(short)*2;
//	waveOutPrepareHeader ( g_wo, &hdr, sizeof(hdr) );
//	waveOutWrite ( g_wo, &hdr, sizeof(hdr) );

FILE *f;
	int size=4*PLAY_TIME*SAMPLE_FREQUENCY;
f= fopen("my.wav","wb");
	wavhdr.Sizeofdata=size;
	wavhdr.size36=size+36;
fwrite(&wavhdr,44,1,f); 
fwrite(finalBuf,size,1,f); 
fclose(f); 


}


/****************************************************************************
 *                                                                          *
 * Function: main                                                           *
 *                                                                          *
 * Purpose : Main entry point.                                              *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

int main(int argc, char *argv[])
{
printf("start making wav\n\r");
    makefile();
    return 0;
}

